; System Module Version 1.1
; Copyright (C) 2011 saguruma
;
; This program is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program.  If not, see <http://www.gnu.org/licenses/>.

; * NOTICE *
; This module requires to locate at the end of PRG-ROM.

; *** I/O Register ***
REG_PPU_0          = $2000 ; PPU #0 (Control) Write only
                           ; bit7 enable NMI (1:enable)
                           ; bit6 always 0 (Master)
                           ; bit5 sprite size (0:8x8 1:8x16)
                           ; bit4 Pattern-VRAM Base Address for BG (0:$0000 1:$1000)
                           ; bit3 Pattern-VRAM Base Address for sprite (0:$0000 1:$1000)
                           ; bit2 Auto Increment mode when VRAM Access (0:1byte 1:32byte)
                           ; bit1-0 Screen Select

REG_PPU_1          = $2001 ; PPU #1 (Control) Write only
                           ; bit7-5 Color Control
                           ; bit4 enable sprite (1:enable)
                           ; bit3 enable BG (1:enable)
                           ; bit2 Left-End mask for sprite (1:enable)
                           ; bit1 Left-End mask for BG (1:enable)
                           ; bit0 Monotone (0:Color 1:Monotone)

REG_PPU_2          = $2002 ; Register PPU #2 (Status) Read only
                           ; bit7 on VBlank (1:VBlank)
                           ; bit6 #0 sprite hit (1:hit)
                           ; bit5 scanline sprite overflow
                           ; bit4 able to Access VRAM
                           ; bit3-0 N/A

REG_PPU_3          = $2003 ; Register PPU #3 (Sprite Address) Write only
REG_PPU_4          = $2004 ; Register PPU #4 (Sprite Access) Write only

REG_PPU_5          = $2005 ; Register PPU #5 (Scroll) Write only (2times)
                           ; first, read ppu status($2002)
                           ; then write scroll values for 2times that Horizontal after Virtical.

REG_PPU_6          = $2006 ; Register PPU #6 (Memory Address) Write only
                           ; write Address to Access VRAM that MSB after LSB.

REG_PPU_7          = $2007 ; Register PPU #7 (Memory Access) Read/Write

REG_DMA            = $4014 ; Sprite DMA Write
                           ; write Address MSB.
REG_PAD_1          = $4016 ; Register PAD #1
REG_PAD_2          = $4017 ; Register PAD #2
REG_MMC_0          = $8000 ; MMC1 #0 (Status)
REG_MMC_1          = $A000 ; MMC1 #1 (CHR bank change)
REG_MMC_2          = $C000 ; MMC1 #2 (CHR bank change)
REG_MMC_3          = $E000 ; MMC1 #3 (PRG bank change)

; *** constants ***
DMA_BASE           = $02   ; DMA Base Address
SPRITE_Y           = $0200 ; Sprite axis value for Y
SPRITE_I           = $0201 ; Sprite tile index value
SPRITE_A           = $0202 ; Sprite attribute
SPRITE_X           = $0203 ; Sprite axis value for X

; *** RAM reserve ***
currentActivity    = $E8   ; current level for activity
work1              = $E9
work2              = $EA
work3              = $EB
work4              = $EC
work5              = $ED
                           ; $EE-$EF is reserved by mod_rng
ppu0               = $F0   ; Copy of Register PPU #0
ppu1               = $F1   ; Copy of Register PPU #1
address1_LSB       = $F2   ; Address #1 LSB
address1_MSB       = $F3   ; Address #1 MSB
address2_LSB       = $F4   ; Address #2 LSB
address2_MSB       = $F5   ; Address #2 MSB
wordparam_LSB      = $F6   ; 16bit work LSB
wordparam_MSB      = $F7   ; 16bit work MSB
pad1               = $F8   ; Pad1 State
pad2               = $F9   ; Pad2 State
pad1prev           = $FA   ; Pad1 Preview State
pad2prev           = $FB   ; Pad2 Preview State
pad1trig           = $FC   ; Pad1 Trigger
pad2trig           = $FD   ; Pad2 Trigger
scrollH            = $FE   ; H-Value of Scroll Register
scrollV            = $FF   ; V-Value of Scroll Register

activities         = $0300 ; state counters what selected activitis
processCounters    = $0308 ; process counters each activities


;******************************************************************************
; RESET
;******************************************************************************
RESET:
  sei ; disable BRK/IRQ
  cld ; disable BCD

  lda #$00
  sta REG_PPU_0

.waitVBlank1:
  bit REG_PPU_2
  bpl .waitVBlank1
.waitScan1:
  bit REG_PPU_2
  bmi .waitScan1
.waitVBlank2:
  bit REG_PPU_2
  bpl .waitVBlank2

  ldx #$00
  stx REG_PPU_0
  stx REG_PPU_1
  dex
  txs

  jsr clearRAM
  jsr clearVRAM

  jsr INIT ; User Initialize

.waitVBlank3:
  bit REG_PPU_2
  bpl .waitVBlank3
.waitScan2:
  bit REG_PPU_2
  bmi .waitScan2
.waitVBlank4:
  bit REG_PPU_2
  bpl .waitVBlank4

  lda <ppu0
  sta REG_PPU_0
  
  lda <ppu1
  sta REG_PPU_1

  cli ; enable BRK/IRQ
LOOP:
  jmp LOOP

;******************************************************************************
; Interrupt: NMI
;******************************************************************************
NMI:
  ; stop NMI
  lda <ppu0
  and #$7B
  sta REG_PPU_0
  
  ; disable BG&SP
  lda <ppu1
  and #$E7
  sta REG_PPU_1

  ; request SP-DMA
  lda #DMA_BASE
  sta REG_DMA

  jsr updatePadState

  jsr MAIN0 ; User Main #0

  ; update scroll register
  lda REG_PPU_2
  lda <scrollH
  sta REG_PPU_5
  lda <scrollV
  sta REG_PPU_5

  lda <ppu0
  sta REG_PPU_0
  lda <ppu1
  sta REG_PPU_1

  jsr MAIN1 ; User Main #1

  rti

;******************************************************************************
; Interrupt: IRQ
;******************************************************************************
IRQ:
  rti

;******************************************************************************
; clear RAM
;******************************************************************************
clearRAM:
  lda #$00
  tax
.loop:
  sta <$00, x
  sta $0200, x
  sta $0300, x
  sta $0400, x
  sta $0500, x
  sta $0600, x
  sta $0700, x
  inx
  bne .loop
  rts

;******************************************************************************
; clear VRAM
;******************************************************************************
clearVRAM:
  ldy #$20
  ldx #$00
  lda REG_PPU_2
  sty REG_PPU_6
  stx REG_PPU_6
  lda #0
.loop:
  sta REG_PPU_7
  inx
  bne .loop
  iny
  cpy #$28
  bne .loop
  rts

;******************************************************************************
; Load VRAM
;     Source Address (MainRAM)    : address1
;     Destinaition Address (VRAM) : address2
;     Size                        : wordparam
;******************************************************************************
loadVRAM:

  lda #0
  tax
  tay
  lda REG_PPU_2
  lda <address2_MSB
  sta REG_PPU_6
  lda <address2_LSB
  sta REG_PPU_6
.loop0:
  lda [address1_LSB], y
  sta REG_PPU_7
.inc1:
  inc <address1_LSB
  bne .inc2
  inc <address1_MSB
.inc2:
  inc <address2_LSB
  bne .decLeftMSB
  inc <address2_MSB
.decLeftMSB:
  lda <wordparam_LSB
  bne .decLeftLSB
  dec <wordparam_MSB
.decLeftLSB:
  dec <wordparam_LSB
  bne .loop0
  lda <wordparam_MSB
  bne .loop0
  rts


;******************************************************************************
; update pad state
;******************************************************************************
updatePadState:
  lda <pad1
  sta <pad1prev
  lda <pad2
  sta <pad2prev

  ldx #$01
  stx REG_PAD_1 ; reset
  dex
  stx REG_PAD_1 ; clear

  ldx #8
.scan:
  lda <pad1
  asl a
  sta <pad1
  lda REG_PAD_1
  and #$01
  ora <pad1
  sta <pad1

  lda <pad2
  asl a
  sta <pad2
  lda REG_PAD_2
  and #$01
  ora <pad2
  sta <pad2
  
  dex
  bne .scan

  lda <pad1
  eor <pad1prev
  and <pad1
  sta <pad1trig

  lda <pad2
  eor <pad2prev
  and <pad2
  sta <pad2trig
  rts

;******************************************************************************
; ResetMMC (Mirroring)
;******************************************************************************
MMC_resetModeMirroring:
  ; reset register
  lda #$80
  sta REG_MMC_0
  sta REG_MMC_1
  sta REG_MMC_2
  sta REG_MMC_3

  lda #%00011110
;          ||||+ Mirroring       : vertical
;          |||+- Screen Mode     : mirroring mode
;          ||+-- PRG Bank Scope  : low($8000)
;          |+--- PRG Bank Size   : 16Kbyte
;          +---- CHR Bank Size   : 4Kbyte

  
  sta REG_MMC_0 ; 5bit write
  lsr a
  sta REG_MMC_0
  lsr a
  sta REG_MMC_0
  lsr a
  sta REG_MMC_0
  lsr a
  sta REG_MMC_0
  nop ; wait
  nop
  rts

;******************************************************************************
; ResetMMC(SingleScreen)
;******************************************************************************
MMC_resetModeSingleScreen:
  ; reset register
  lda #$80
  sta REG_MMC_0
  sta REG_MMC_1
  sta REG_MMC_2
  sta REG_MMC_3

  lda #%00011100
;          ||||+ Mirroring       : vertical
;          |||+- Screen Mode     : single-screen mode
;          ||+-- PRG Bank Scope  : low($8000)
;          |+--- PRG Bank Size   : 16Kbyte
;          +---- CHR Bank Size   : 4Kbyte
  
  sta REG_MMC_0 ; 5bit write
  lsr a
  sta REG_MMC_0
  lsr a
  sta REG_MMC_0
  lsr a
  sta REG_MMC_0
  lsr a
  sta REG_MMC_0
  nop ; wait
  nop
  rts

;******************************************************************************
; ChangePRG
;******************************************************************************
MMC_changePRG:
  sta REG_MMC_3 ; 5bit write
  lsr a
  sta REG_MMC_3
  lsr a
  sta REG_MMC_3
  lsr a
  sta REG_MMC_3
  lsr a
  sta REG_MMC_3
  nop ; wait
  nop
  rts

;******************************************************************************
; ChangeCHR (Low4Kbyte)
;******************************************************************************
MMC_changeCHR_Low:
  sta REG_MMC_1 ; 5bit write
  lsr a
  sta REG_MMC_1
  lsr a
  sta REG_MMC_1
  lsr a
  sta REG_MMC_1
  lsr a
  sta REG_MMC_1
  nop ; wait
  nop
  rts

;******************************************************************************
; ChangeCHR (High4Kbyte)
;******************************************************************************
MMC_changeCHR_High:
  sta REG_MMC_2 ; 5bit write
  lsr a
  sta REG_MMC_2
  lsr a
  sta REG_MMC_2
  lsr a
  sta REG_MMC_2
  lsr a
  sta REG_MMC_2
  nop ; wait
  nop
  rts

;******************************************************************************
; Interrupt Vector
;******************************************************************************
  .org $FFFA
  .dw NMI
  .dw RESET
  .dw IRQ
